// QUBACQNM.cpp : 
//		Acquisition card interface for National Instruments mx Library
//
// Note : For AI channel create, it seems the minimum and maximum values are 
// used to determine the largest scaling factor which will include the range 
// and then it is automatically set.  Somewhere, scaling factor actually 
// applied should be read back in.  The DAQmxSetAIGain/DAQmxGetAIGain however
// seem to cause a 'not available in task' error wherever they are call from.
// The NI doc's indicate there are several ways to set the gain, including 
// using the min/max values, but no explanation on how to read which gain 
// was actually selected when the min and max lie between gain settings.
// For now, assume the user will set gain to a reasonable value (1,2,10,?)
// and set the min/max accordingly which then causes the gain to be what the 
// user desired.
//
// TODO : Need to test this driver with the E series cards - the default range
// on those is +-5 volts?  Gain settings of -1 for .5 ? 
//
// TODO : Test with 18 bit data card (possibly a simulated card)
//
#include "stdafx.h"
#define QUBDLL extern "C" __declspec(dllexport)
#include "NIDAQmx.h"
#include "stdlib.h"
#include "stdio.h"
#include "process.h"
#include <shlobj.h>

#include <fstream>

//========================================================= STRUCTURES 
#pragma pack(push, 1)
struct InputStart_Params {
	long iSize;
	long iChannelCount;
	long * aiChannels;
	double * anGain;
	long iBits;
	double nSampling;
	void (__stdcall *pfnCallBack ) (void * pCaller, void * pBuffer, long iBufferSize, long iBufferCount, long iError);
	void * pCaller;				//For pass back in pfnCallBack
	long iBufferLength;
	void * pBuffer;				//Is filled by the driver with the address of the internal buffer
	BYTE bSingleBuffer;			
	long iBufferCount; 
	long iBufferElemSize;		//Filled by the driver
	void * pDeviceData;			//Pointer to a hardware-specific structure, Allocated/freed by the QUBDAQ driver Start/Stop functions
	};

struct OutputStart_Params {
	long iSize;
	long iChannelCount;
	long iDigitalChannelCount;
	long * aiChannels;
	double * anGain;
	long iBits;
	double nSampling;
	double nDigitalSampling;
	void (__stdcall *pfnCallBack) (void * pCaller, void * pBuffer, long iBufferSize, long * piBufferCount, long * piError);
	void (__stdcall *pfnDigCallBack) (void * pCaller, long *piDigVal, long *piDigCount, long * piError);
	void * pCaller;				//For pass back in pfnCallBack
	long iBufferLength;
	void * pBuffer;				//Is filled by the driver with the address of the internal buffer
	BYTE bSingleBuffer;			
	long iBufferCount; 
	long iBufferElemSize;		//Filled by the driver
	void * pDeviceData;			//Pointer to a hardware-specific structure, Allocated/freed by the QUBDAQ driver Start/Stop functions
	};

#pragma pack(pop)					

// #define RAWBUFFER

//========================================================= GLOBALS
char * g_pchName="QUB NI-MX 1.0";		// DLL reference name
const bool32 AUTOSTART = TRUE;			// named const for DAQmxWrite functions 

long g_iError;

//----- Device names
int g_iDeviceCount;						// Number of Ni-MX devices detected.
const int ACQ_MAX_DEVICES = 16;			// Number of devices to create name array
const int MAXNAMELEN = 64;				// Length of each name in name array 
BYTE g_achDeviceName[ACQ_MAX_DEVICES][MAXNAMELEN];	

//----- three tasks - multiple simultaneous input or output tasks not currently supported.
TaskHandle g_hIn=NULL;			// DaqMX task handle for input task 
TaskHandle g_hOut=NULL;			// DaqMX task handle for Output task 

InputStart_Params g_isp;		// Parameters in use for current input operation
OutputStart_Params g_osp;		// Parameters in use for current Output operation
short * g_aiOutBuf=NULL;		// Buffer passed to QUB to store data to be written.
short * g_aiInBuf=NULL;			// Data copied from driver, passed back to QUB.

const int INPUTBUFFERS = 4;		// Size of NI buffer as multiplier to size of buffer exchanged with Qub.

#ifdef RAWBUFFER 

	//----- In debug mode, a text file is written to for DAQmx calls.
	const int INITRAWBUFFERS = 16;		// Buffers to hold data waiting to be sent to QUB
	const int MAXRAWBUFFERS = 100;		// max buffers for acquisition to lag QuB.
	FILE * g_hAcqRaw=NULL;				// Raw buffer data logging 
	short * g_aiInRaw[MAXRAWBUFFERS];	// Data copied from driver, passed back to QUB.
	int g_alFlagRaw[MAXRAWBUFFERS];		// 0=empty 1=data ready to be Moved to Qub
	int g_iRawBuffers;					// Number of allocated buffers 

	HANDLE g_hRawEvent;					// Event handle for notifying thread that a new raw buffer is available
	HANDLE g_hRawThread;				// Thread handle for thread which processes raw buffers to file and Qub
	bool g_lRawExit;					// Used to notify thread of desire to quit 

	// Add another buffer - true success, false - failed ( too many or out of memory ) 
	bool CreateRawBuffer( ) { 
		if( g_iRawBuffers >= MAXRAWBUFFERS ) 
			return false;
		if( (g_aiInRaw[g_iRawBuffers]=new short[g_isp.iBufferCount*g_isp.iChannelCount]) == NULL ) 
			return false;
		g_alFlagRaw[g_iRawBuffers]=0;
		g_iRawBuffers++;	// must increment after setting flag so it is ready to go immediately 
		return true;
		}

	// Find a not used buffer or attempt to create one if all are in use.
	// TODO - check for -1 return -- error in create buffer.... 
	int FindOrCreateRawBuffer( ) { 
		for( int i=0; i<g_iRawBuffers; ++i ) 
			if( g_alFlagRaw[i]==0 ) 
				return i;
		if( CreateRawBuffer() ) {
			return g_iRawBuffers-1;
			}
		return -1;
		}

	// Initialize the buffers
	void CreateRawBuffers( ) { 
		g_hAcqRaw = fopen("C:\\QubAcq.Raw", "wb");
		int iBuf;
		for( iBuf=0; iBuf<MAXRAWBUFFERS; ++iBuf ) 
			g_aiInRaw[iBuf]=NULL;

		g_iRawBuffers=0;
		for( iBuf=0; iBuf<INITRAWBUFFERS; ++iBuf )
			CreateRawBuffer();
		}

	// Clear the raw buffers memory
	void DeleteRawBuffers(){ 
		while( g_iRawBuffers > 0 ) 
			delete [] g_aiInRaw[--g_iRawBuffers];
		fclose( g_hAcqRaw );
		}

#endif

//========================================================= DLL Entry Point
BOOL APIENTRY DllMain( HANDLE hInstance, DWORD  dwReason, LPVOID lpReserved ){
    switch( dwReason ){
		case DLL_PROCESS_ATTACH:
			break;
		case DLL_THREAD_ATTACH:
			break;
		case DLL_THREAD_DETACH:
			break;
		case DLL_PROCESS_DETACH:
			break;
		}
    return TRUE;
	}

//====================================================== Library Setup Functions
//----- Library Initialize / Terminate
QUBDLL long __stdcall LibraryInitialize(){
	//----- Fill device count and array of device names : g_achDeviceName, g_iDeviceCount
	g_iDeviceCount=0;
	char buf[ACQ_MAX_DEVICES*MAXNAMELEN];
	g_iError=DAQmxGetSysDevNames(buf, ACQ_MAX_DEVICES*MAXNAMELEN);

	//----- parse string of names to array of strings (pseudo strtok)
	if( g_iError==0 && buf[0]!=0 ) { 
		int iArrayPos=0;
		for( int iPos=0; iPos<ACQ_MAX_DEVICES*MAXNAMELEN && g_iDeviceCount<ACQ_MAX_DEVICES; ++iPos) { 
			if( buf[iPos]==0 ) {
				// end of string : exit 
				++g_iDeviceCount;
				break;
				}
			else if( buf[iPos]==',' ) {
				// Start next string 
				g_achDeviceName[g_iDeviceCount][iArrayPos]=0;
				iArrayPos=0;
				++g_iDeviceCount;
				if( buf[iPos+1]==' ')
					++iPos;
				}
			else if( iArrayPos+1<MAXNAMELEN )
				// copy one char from buffer to name array 
				g_achDeviceName[g_iDeviceCount][iArrayPos++]=buf[iPos];
			}
		}

	return 0;
	}

QUBDLL long __stdcall LibraryTerminate(){
	return 0;
	}

//----- Copy Library name to provided buffer
QUBDLL long __stdcall LibraryNameLength(){
	return strlen(g_pchName)+1;
	}

QUBDLL void __stdcall LibraryName( char * pchName ){
	strcpy(pchName,g_pchName);
	}

//========================================================= Single Sample Input

//----- TODO - used only by KPL ? 
//.PAS : Result := AI_VRead(Devices[Device], Channel, Round(Gain), @Value);
QUBDLL long __stdcall ReadDBLChannel(long iDevice, long iChannel, double nGain, double * pnValue){
	*pnValue=0.0;
	return 0;
	}

//----- Analog Input - Single sample, Single Channel
// int32 DllExport __CFUNC DAQmxSetAIGain(TaskHandle taskHandle, const char channel[], float64 data);
QUBDLL long __stdcall AIRead(long iDevice, long iChannel, double nGain, long * piValue){
	short aiInpData[2];				// Data storage (2 samples to accomodate 18 bit data)
	long iInpSampCount;				// result - number of samples read 
	long iInpBytesPerSamp;			

	//-- Create channel name string
	char achChannelName[32];
	sprintf(achChannelName, "%s/ai%d", g_achDeviceName[iDevice], iChannel );

	//-- Create Task
	TaskHandle h;
	if( (g_iError=DAQmxCreateTask( NULL, &h )) != 0 ){
		return g_iError;
		}

	//-- Add channel 
	if( (g_iError=DAQmxCreateAIVoltageChan( h, achChannelName, NULL, DAQmx_Val_RSE, -10.0, 10.0, DAQmx_Val_Volts, NULL) )!=0 ) {
		DAQmxClearTask( h );
		return g_iError;
		}

	//-- Start Task 
	if( (g_iError=DAQmxStartTask( h )) != 0 ){
		DAQmxClearTask( h );
		return g_iError;
		}

	//-- Read data  & Check 
	if( (g_iError=DAQmxReadRaw( h, 1/*Samples*/, 1.0 /*Try 1 second*/, aiInpData, 2*sizeof(short)/*Arraysize*/, &iInpSampCount, &iInpBytesPerSamp, NULL) != 0 )){
		DAQmxClearTask( h );
		return g_iError;
		}
	if( iInpSampCount<1 ){
		DAQmxClearTask( h );
		return -1;
		}
	if( iInpBytesPerSamp>2 )		// Check for 18 bit output and convert.
		aiInpData[0]= short((*((long *)aiInpData))>>2);

	//-- Stop Task 
	DAQmxStopTask( h );
	DAQmxClearTask( h );

	*piValue = aiInpData[0];
	return 0;
	}

//----- Digital Input - single sample, single channel
QUBDLL long __stdcall DIRead( long iDevice, long iChannel, long * piValue ){
	BYTE abData[2]={0,0};
	long iInpSampCount;				// result - number of samples read 
	long iInpBytesPerSamp;			

	//-- Create channel name string  ex. "Dev0/line0"
	char achChannelName[32];
	sprintf(achChannelName, "%s/line%d", g_achDeviceName[iDevice], iChannel );

	//-- Create Task
	TaskHandle h;
	if( (g_iError=DAQmxCreateTask( NULL, &h )) != 0 )
		return g_iError;

	//-- Add channel 
	if( (g_iError=DAQmxCreateDIChan( h, achChannelName, NULL, DAQmx_Val_ChanForAllLines) )!=0 ) {
		DAQmxClearTask( h );
		return g_iError;
		}
	
	//-- Start Task 
	if( (g_iError=DAQmxStartTask( h )) != 0 ){
		DAQmxClearTask( h );
		return g_iError;
		}

	//-- Read data  & Check 
	if( (g_iError=DAQmxReadDigitalLines( h, 1/*Samples*/, 1.0 /*Try once*/, DAQmx_Val_GroupByChannel, abData, 2/*Arraysize*/, &iInpSampCount, &iInpBytesPerSamp, NULL) != 0 )){
		DAQmxClearTask( h );
		return g_iError;
		}
	if( iInpSampCount<1 ){
		DAQmxClearTask( h );
		return -1;
		}
	
	//-- Stop Task 
	DAQmxStopTask( h );
	DAQmxClearTask( h );

	*piValue=abData[0];
	return 0;
	}

//========================================================= Single Sample Output

//----- Analog Output - Single sample, Single Channel
//----- TODO - used only by KPL ? 
QUBDLL long __stdcall WriteDBLChannel( long iDevice, long iChannel, const double nValue ){
	return 0;
	}

//----- Analog Output - Single Channel, Single Sample 
QUBDLL long __stdcall AOWrite( long iDevice, long iChannel, long iValue ){
	//-- Create channel name string
	char achChannelName[32];
	sprintf(achChannelName, "%s/ao%d", g_achDeviceName[iDevice], iChannel );

	//-- Create Task
	TaskHandle h;
	if( (g_iError=DAQmxCreateTask( NULL, &h )) != 0 )
		return g_iError;

	//-- Add channel 
	if( (g_iError=DAQmxCreateAOVoltageChan( h, achChannelName, NULL,  -10.0, 10.0, DAQmx_Val_Volts, NULL) )!=0 ) {
		DAQmxClearTask( h );
		return g_iError;
		}
	
	//-- Start Task 
	if( (g_iError=DAQmxStartTask( h )) != 0 ){
		DAQmxClearTask( h );
		return g_iError;
		}

	//-- Read data  & Check 
	short iOutData=(short)iValue;
	int32 iOutSampCount;
	if( (g_iError=DAQmxWriteRaw( h, 1/*Samples*/, AUTOSTART, 0.0 /*Try once*/, &iOutData, &iOutSampCount, NULL) != 0 )){
		DAQmxClearTask( h );
		return g_iError;
		}
	if( iOutSampCount<1 ){
		DAQmxClearTask( h );
		return -1;
		}

	//-- Stop Task 
	DAQmxStopTask( h );
	DAQmxClearTask( h );

	return 0;
	}

//----- Digital Output - Single Channel, Single Sample 
QUBDLL long __stdcall DOWrite( long iDevice, long iChannel, const long iValue ){
	BYTE abData[4]={(iValue==0?0:1),0,0,0};

	//-- Create channel name string  ex. "Dev0/line0"
	char achChannelName[32];
	sprintf(achChannelName, "%s/line%d", g_achDeviceName[iDevice], iChannel );

	//-- Create Task
	TaskHandle h;
	if( (g_iError=DAQmxCreateTask( NULL, &h )) != 0 )
		return g_iError;

	//-- Add channel 
	if( (g_iError=DAQmxCreateDOChan( h, achChannelName, NULL, DAQmx_Val_ChanForAllLines) )!=0 ) {
		DAQmxClearTask( h );
		return g_iError;
		}
	
	//-- Start Task 
	if( (g_iError=DAQmxStartTask( h )) != 0 ){
		DAQmxClearTask( h );
		return g_iError;
		}

	//-- Write Data 
	int32 iWritten;
	if( (g_iError=DAQmxWriteDigitalLines( h, 1/*Samples*/, AUTOSTART, 0.0/*TryOnce*/, DAQmx_Val_GroupByChannel, abData, &iWritten, NULL)) != 0 
			|| iWritten<1 ){
		DAQmxClearTask( h );
		return -1;
		}
	
	//-- Stop Task 
	DAQmxStopTask( h );
	DAQmxClearTask( h );

	return 0;
	}

//---- Output callback to refill generated data buffer
// this is the [N.I.->dll] callback which makes the [dll->Qub] callback occur.
long __cdecl EveryNSamplesOutputCallback( TaskHandle hTask, long everyNsamplesEventType, uInt32 nSamples, void *pvoid ){
	long iWrittenCount=nSamples, iCopied=nSamples;
	int32 iError=0;
	g_osp.pfnCallBack( g_osp.pCaller, g_aiOutBuf, g_osp.iBufferElemSize, &iWrittenCount, &iError );
	iError=DAQmxWriteBinaryI16( g_hOut, iWrittenCount, 0, 10.0, DAQmx_Val_GroupByScanNumber, g_aiOutBuf, &iCopied, NULL );
	return iError;
	}

//---- Input callback to copy acquired data 
// this is the [N.I.->dll] callback which makes the [dll->Qub] callback occur.
long __cdecl EveryNSamplesInputCallback(TaskHandle hTask, long everyNsamplesEventType, uInt32 nSamples, void *pvoid){
	long iReadCount;
	int32 iError=0;
	iError=DAQmxReadBinaryI16( g_hIn, g_isp.iBufferCount, 10.0, DAQmx_Val_GroupByScanNumber, g_aiInBuf, g_isp.iBufferCount*g_isp.iChannelCount, &iReadCount, NULL);
	g_isp.pfnCallBack( g_isp.pCaller, g_aiInBuf, g_isp.iBufferElemSize, iReadCount, iError );

	return iError;
	}

//---- Experiment 11/05 - quickly copy the raw data to a buffer and return.
// use an additional thread to process the data.
#ifdef RAWBUFFER 
	long __cdecl EveryNSamplesRawCallback(TaskHandle hTask, long everyNsamplesEventType, uInt32 nSamples, void *pvoid){
		int iBuf = FindOrCreateRawBuffer();
		if( iBuf == -1 ){ 
			//TODO - ?? 
			return -1;
			}
		
		long iReadCount;
		int32 iError=0;
		iError=DAQmxReadBinaryI16( g_hIn, g_isp.iBufferCount, 10.0, DAQmx_Val_GroupByScanNumber, g_aiInRaw[iBuf], g_isp.iBufferCount*g_isp.iChannelCount, &iReadCount, NULL);
		g_alFlagRaw[iBuf]=1;	// set AFTER fill in case thread processing prior buffer finds this one too.
		//Signal Raw data mover thread that data is ready. (If already set, can safely set again.)
		SetEvent( g_hRawEvent );		
		return iError;
		}

	// Raw data mover Thread Function 
	VOID DataMover(LPVOID lpParam) { 
		int iBuf, iError=0;
		while( WaitForSingleObject(g_hRawEvent,INFINITE)== WAIT_OBJECT_0 && ! g_lRawExit ) { 
			//-- Process all waiting buffers (possibly none if caught during prior signal processing)
			for( iBuf=0; iBuf<g_iRawBuffers; ++iBuf ) {
				if( g_alFlagRaw[iBuf] ) { 
					//-- Process one buffer (iBuf) : copy to file 
					fwrite( g_aiInRaw[iBuf], g_isp.iBufferElemSize, g_isp.iBufferCount, g_hAcqRaw );

					//-- : Copy buffer to Qub
					g_isp.pfnCallBack( g_isp.pCaller, g_aiInRaw[iBuf], g_isp.iBufferElemSize, g_isp.iBufferCount, iError );
					//todo - iError handling

					//-- Make buffer available again 
        			g_alFlagRaw[iBuf]=0;
					}
				}
			}

		//-- WAIT_ABANDONED | WAIT_TIMEOUT | g_lRawExit=true
		CloseHandle(g_hRawEvent);
		}
#endif

//----- Synchronized Input / Output, or separate
QUBDLL long __stdcall SynchStart( long iInputDevice, long iOutputDevice, 
			InputStart_Params * pInputParams, OutputStart_Params * pOutputParams, 
			int * piInputStarted, int * piOutputStarted){
	
	char achChannelName[128];		// temporary buffer for constructing channel names 
	int32 iError=0;					// Result from last mx Call 

	CHAR my_documents[MAX_PATH];
	HRESULT doc_result = SHGetFolderPath(NULL, CSIDL_PERSONAL, NULL, SHGFP_TYPE_CURRENT, my_documents);
	sprintf(my_documents + strlen(my_documents), "\\qubacqnm_log.txt");
	std::ofstream dbg_out;
	dbg_out.open(my_documents, std::ios_base::out);

	//----- Analog Output Setup
	g_osp.iChannelCount=0;			// pre-set ChannelCount to 0, in case OutParams is NULL
	g_hOut=NULL;
	if( pOutputParams!=NULL && pOutputParams->iChannelCount>0 && iError==0 ) { 
		g_osp = * pOutputParams;

		//-- Create Output Task 
		iError=DAQmxCreateTask( NULL, &g_hOut );
		
		//-- Create analog Output Channels
		for( int iChOut=0; iChOut<g_osp.iChannelCount && iError==0; ++iChOut ) {
			sprintf( achChannelName, "%s/ao%d", g_achDeviceName[iOutputDevice], g_osp.aiChannels[iChOut] );
			iError=DAQmxCreateAOVoltageChan( g_hOut, achChannelName, NULL, -10.0, 10, DAQmx_Val_Volts, NULL);
			}

		//-- ensure regen mode is off for continuous generation
		if( iError==0 ) 
			iError=DAQmxSetWriteRegenMode( g_hOut, DAQmx_Val_DoNotAllowRegen );
		
		//-- Set on card buffer size to 2x qub requested buffer size (overridden by SampClkTiming ?)
		if( iError==0 ) {
			iError=DAQmxCfgOutputBuffer( g_hOut, 2*g_osp.iBufferLength/*Samples*/ );
			}

		//-- set sampling rate (g_osp.nSampling) and buffer size (g_osp.iBufferLength*g_osp.iChannelCount)
		if( iError==0 ) {
			iError=DAQmxCfgSampClkTiming( g_hOut, NULL, g_osp.nSampling, DAQmx_Val_Rising, DAQmx_Val_ContSamps, 2*g_osp.iBufferLength);
			}

		//-- Set up a callback function to occur after every g_osp.iBufferLength samples are written
		iError=DAQmxRegisterEveryNSamplesEvent( g_hOut, DAQmx_Val_Transferred_From_Buffer, g_osp.iBufferLength, 0, EveryNSamplesOutputCallback, NULL);

		//-- Create buffer for exchanging data with QUB
		g_aiOutBuf = new short[g_osp.iBufferLength*g_osp.iChannelCount];
		if( g_aiOutBuf == NULL )
			iError=-1001;
		g_osp.iBufferElemSize = g_osp.iChannelCount * sizeof(short);

		//-- fill Buffer 
		EveryNSamplesOutputCallback( g_hOut, -1, g_osp.iBufferLength, NULL );
		EveryNSamplesOutputCallback( g_hOut, -1, g_osp.iBufferLength, NULL );
		}

	//----- Analog Input Setup
	g_isp.iChannelCount=0;			// pre-set ChannelCount to 0, in case InParams is NULL
	g_hIn=NULL;
	if( pInputParams!=NULL && pInputParams->iChannelCount > 0 && iError==0 ) { 
		g_isp = * pInputParams;
		//-- Create Input Task 
		iError=DAQmxCreateTask( NULL, &g_hIn );

		//-- Create Analog Input Channels	//TODO : use 5.0/Gain for e Series cards.
		for( int iChIn=0; iChIn<g_isp.iChannelCount && iError==0; ++iChIn ) {
			sprintf( achChannelName, "%s/ai%d", g_achDeviceName[iInputDevice], g_isp.aiChannels[iChIn] );
			dbg_out << "Creating input channel " << achChannelName << "..." << std::endl;
			iError=DAQmxCreateAIVoltageChan( g_hIn, achChannelName, NULL, DAQmx_Val_RSE, -10.0/g_isp.anGain[iChIn], 10.0/g_isp.anGain[iChIn], DAQmx_Val_Volts, NULL);
			if ( iError ) {
				char errorString[10240];
				DAQmxGetErrorString(iError, errorString, 10240);
				dbg_out << "error " << iError << ": " << errorString << std::endl;
				DAQmxGetExtendedErrorInfo(errorString, 10240);
				dbg_out << errorString << std::endl;
				dbg_out << "Trying differential config..." << std::endl;

				iError=DAQmxCreateAIVoltageChan( g_hIn, achChannelName, NULL, DAQmx_Val_Diff, -10.0/g_isp.anGain[iChIn], 10.0/g_isp.anGain[iChIn], DAQmx_Val_Volts, NULL);
				if ( iError ) {
					char errorString[10240];
					DAQmxGetErrorString(iError, errorString, 10240);
					dbg_out << "error " << iError << ": " << errorString << std::endl;
					DAQmxGetExtendedErrorInfo(errorString, 10240);
					dbg_out << errorString << std::endl;
				} else {
					dbg_out << "success." << std::endl;
				}
			} else {
				dbg_out << "success." << std::endl;
			}
		}

		//-- Set buffer size - if 0, use the same logic as QUB
		if( g_isp.iBufferCount==0 ) {
			g_isp.iBufferCount = (int(g_isp.nSampling*g_isp.iBufferLength/1000+0.5)+511)/512;
			g_isp.iBufferCount = (g_isp.iBufferCount>2048?2048:g_isp.iBufferCount)*512;
			}
    
		//-- Write buffer size to task 
		g_isp.iBufferElemSize = g_isp.iChannelCount * sizeof(short);
		iError=DAQmxCfgInputBuffer( g_hIn, INPUTBUFFERS*g_isp.iBufferCount );
		g_aiInBuf = new short[g_isp.iBufferCount*g_isp.iChannelCount];
		if( g_aiInBuf == NULL )
			iError=-1001;

		//-- if simultaneous Acq and Gen, synchronized Acq with Gen, otherwise set up Acq Clock.
		if( g_osp.iChannelCount ) { 
			sprintf( achChannelName, "/%s/ao/SampleClock", g_achDeviceName[iOutputDevice], g_isp.aiChannels[iChIn] );
			iError=DAQmxCfgSampClkTiming( g_hIn, achChannelName, g_isp.nSampling, DAQmx_Val_Rising, DAQmx_Val_ContSamps, 2*g_osp.iBufferLength /*g_isp.iBufferLength*/ );
			}
		else 
			iError=DAQmxCfgSampClkTiming( g_hIn, NULL, g_isp.nSampling, DAQmx_Val_Rising, DAQmx_Val_ContSamps, 2*g_isp.iBufferLength );

		#ifdef RAWBUFFER 
			//-- Create Event ( security, manual reset, intially signaled, name )
			if( (g_hRawEvent = CreateEvent( NULL, FALSE, FALSE, "BUFFERFILLED" )) == NULL ) { 
				//todo - error
				}

			g_lRawExit = false;
			DWORD dwID;
			if( (g_hRawThread = CreateThread( NULL, 0, (LPTHREAD_START_ROUTINE) DataMover, NULL, 0, &dwID)) == NULL ) { 
				//todo - error
				}

			//-- Set up a callback function to occur after every g_isp.iBufferCount samples are written		// DAQmx_Val_SynchronousEventCallbacks 
			DAQmxRegisterEveryNSamplesEvent( g_hIn, DAQmx_Val_Acquired_Into_Buffer, g_isp.iBufferCount, 0, EveryNSamplesRawCallback, NULL);

			//-- Set up buffers and raw file 
			CreateRawBuffers( );
		#else
			//-- Set up a callback function to occur after every g_isp.iBufferCount samples are written
			DAQmxRegisterEveryNSamplesEvent( g_hIn, DAQmx_Val_Acquired_Into_Buffer, g_isp.iBufferCount, 0, EveryNSamplesInputCallback, NULL);
		#endif
		}

	//----- Set AI to trigger on start of AO 
	//if ( g_osp.iChannelCount>0 && g_isp.iChannelCount>0 ) {
	//	sprintf( achChannelName, "/%s/ao/StartTrigger", g_achDeviceName[iOutputDevice] );
	//	iError=DAQmxCfgDigEdgeStartTrig( g_hIn, achChannelName, DAQmx_Val_Falling );
	//	}

	//-- Start Task(s)
	if( iError==0 && g_isp.iChannelCount>0 ) {
		iError=DAQmxStartTask(g_hIn);
		*piInputStarted=(iError==0?1:0);
		}
	else 
		*piInputStarted=0;

	if( iError==0 && g_osp.iChannelCount>0 ) {
		iError=DAQmxStartTask(g_hOut);
		*piOutputStarted=(iError==0?1:0);
		}
	else
		*piOutputStarted=0;

	return iError;
	}

//----- Call SynchStart for input only
QUBDLL long __stdcall InputStart( long iDevice, InputStart_Params * pInputParams ){
	int iIn, iOut;
	return SynchStart( iDevice, -1, pInputParams, NULL, &iIn, &iOut );
	}

//----- Call SynchStart for Output only 
QUBDLL long __stdcall OutputStart( long iDevice, OutputStart_Params * pOutputParams){
	int iIn, iOut;
	return SynchStart( -1, iDevice, NULL, pOutputParams, &iIn, &iOut );
	}

//----- Stop and clear tasks 
void AcquisitionStop(bool lInputStop, bool lOutputStop) { 
	if( lOutputStop ) {
		if( g_hOut!=NULL ) { 
			DAQmxStopTask(g_hOut);
			DAQmxClearTask(g_hOut);
			g_hOut=NULL;
			delete [] g_aiOutBuf;
			}			
		}			

	if( lInputStop ) { 
		if( g_hIn!=NULL ) { 
			DAQmxStopTask(g_hIn);
			DAQmxClearTask(g_hIn);
			g_hIn=NULL;
			delete [] g_aiInBuf;

			#ifdef RAWBUFFER 
				g_lRawExit=true;
				SetEvent( g_hRawEvent );		
				WaitForSingleObject( g_hRawThread, 10000 );	// Wait up to 10 sec. while data streams to Qub
				DeleteRawBuffers();
			#endif
			}			
		}
	}

//----- Stop Acquisition and generation 
QUBDLL long __stdcall SynchStop( ){
	AcquisitionStop( true, true );
	return 0;
	}

//----- Stop Acquisition only
QUBDLL long __stdcall InputStop( long iDevice ){
	AcquisitionStop( true, false );
	return 0;
	}

//----- Stop Generation Only
QUBDLL long __stdcall OutputStop( long iDevice){
	AcquisitionStop( false, true );
	return 0;
	}

//========================================================= Device Properties 
//----- # of devices detected, same for input / output.
QUBDLL long __stdcall InputDeviceCount( ){
	return g_iDeviceCount;
	}
QUBDLL long __stdcall OutputDeviceCount( ){
	return g_iDeviceCount;
	}

//----- Copy device name (ex: Dev1 Dev2 ) to provided buffer 
void CopyDeviceName( long iDevice, char * pchName ) { 
	if(iDevice<0 || iDevice >= g_iDeviceCount )
		pchName[0]=0;
	else
		strcpy(pchName,(const char * )(g_achDeviceName[iDevice]));
	}

QUBDLL void __stdcall InputDeviceName( long iDevice, char * pchName ){
	CopyDeviceName(iDevice,pchName);
	}	

QUBDLL void __stdcall OutputDeviceName( long iDevice, char * pchName ){
	CopyDeviceName(iDevice, pchName);
	} 

//----- Basic properties
QUBDLL long __stdcall InputChannelCount( long iDevice ){		//----- TODO - read from driver
	return 32;
	} 

QUBDLL double __stdcall InputRange( long iDevice ){			//----- TODO - read from driver
	return 20.0;
	} 

QUBDLL double __stdcall InputGain( long iDevice, long iChannel ){	//----- TODO - read from driver
	return 1.0;
	}

QUBDLL long __stdcall InputDifferential( long iDevice ){		//----- TODO - read from driver
	return 1;	// TRUE
	}

QUBDLL long __stdcall InputBits( long iDevice ){				//----- TODO - read from driver
	return 16;
	} 

QUBDLL long __stdcall InputDeviceCode( long iDevice ){		
	//-- Scan device name (ex: PCI-6229) for the first numeric char and 
	//-- atoi() the remainder of the string to get a device code.
	//-- This is used in the QUB hardware setup to initialize the hardware 
	//-- configuration settings.
	char achDeviceName[100];
	g_iError=DAQmxGetDevProductType( (char *)(g_achDeviceName[iDevice]), achDeviceName, 100);
	long iDevCode=0;

	for( unsigned int iChr=0; iChr<strlen((char *)achDeviceName); ++iChr ) 
		if( achDeviceName[iChr]>='0' && achDeviceName[iChr]<='9' ) {
			iDevCode=atoi((char *)(achDeviceName+iChr));
			break;
			}

	return iDevCode;
	}

QUBDLL long __stdcall OutputChannelCount( long iDevice ){	//----- TODO - read from driver
	return 16;
	}

QUBDLL long __stdcall OutputDigitalChannelCount( long iDevice ){	//----- TODO - read from driver
	return 6;
	}

QUBDLL double __stdcall OutputRange( long iDevice ){		//----- TODO - read from driver
	return 20.0;
	}

QUBDLL double __stdcall OutputGain( long iDevice, long iChannel ){//----- TODO - read from driver
	return 1.0;
	}

QUBDLL long __stdcall OutputDifferential( long iDevice ){	//----- TODO - read from driver
	return 1;	// TRUE
	}

QUBDLL long __stdcall OutputBits( long iDevice ){			//----- TODO - read from driver
	return 16;
	}
